/*
    $Id: dimm.h 115 2010-03-14 08:48:08Z matwey $
    Copyright (C) 2010  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _DIMM_H
#define _DIMM_H

#include "table_storage.h"

#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/numeric/ublas/vector.hpp>

namespace dimm {
	using namespace boost::numeric::ublas;

	typedef boost::posix_time::ptime time_type;
	
	typedef double                 value_type;
	typedef vector<value_type>     vector_type;
	typedef vector_type::size_type size_type;

	typedef vector_type dimm_flux_type;
	typedef vector_type dimm_vars_type;

	dimm_vars_type correction( const dimm_vars_type& vars );
	value_type turbulence( value_type scale, value_type Knumber, value_type variance );
	value_type seeing( value_type turbulence_power );

	struct dimm_sample {
		time_type      timestamp;
		unsigned int   framenum;
		dimm_flux_type left_flux;
		dimm_flux_type right_flux;
		dimm_vars_type xvars;
		dimm_vars_type yvars;

		dimm_sample( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars ):
			timestamp(time), framenum(framenum), left_flux(left), right_flux(right), xvars(xvars), yvars(yvars) {}
	};

	const dimm_flux_type& left_flux( const dimm_sample& x );
	const dimm_flux_type& right_flux( const dimm_sample& x );
	const dimm_vars_type& xvars( const dimm_sample& x );
	const dimm_vars_type& yvars( const dimm_sample& x );
};

class dimm_storage:
	public table_storage< dimm::dimm_sample > {
public:
	typedef table_storage_column< dimm::dimm_sample, boost::function1< dimm::dimm_flux_type, dimm::dimm_sample > > left_flux_column_type;
	typedef table_storage_column< dimm::dimm_sample, boost::function1< dimm::dimm_flux_type, dimm::dimm_sample > > right_flux_column_type;
	typedef table_storage_column< dimm::dimm_sample, boost::function1< dimm::dimm_vars_type, dimm::dimm_sample > > xvars_column_type;
	typedef table_storage_column< dimm::dimm_sample, boost::function1< dimm::dimm_vars_type, dimm::dimm_sample > > yvars_column_type;
private:
	left_flux_column_type  m_left_flux_column;
	right_flux_column_type m_right_flux_column;
	xvars_column_type      m_xvars_column;
	yvars_column_type      m_yvars_column;
public:
	dimm_storage( size_type reserve = 60 );

	const left_flux_column_type&  left_flux()  const { return m_left_flux_column; }
	const right_flux_column_type& right_flux() const { return m_right_flux_column; }
	const xvars_column_type&      xvars()      const { return m_xvars_column; }
	const yvars_column_type&      yvars()      const { return m_yvars_column; }

	void push_back( const dimm::time_type& time, unsigned int framenum, const dimm::dimm_flux_type& left, const dimm::dimm_flux_type& right, const dimm::dimm_vars_type& xvars, const dimm::dimm_flux_type& yvars ) {
		// FIXME: dimm_sample_ref
		table_storage< dimm::dimm_sample >::push_back( dimm::dimm_sample( time, framenum, left, right, xvars, yvars ) );
	}

	void correct_low_vars();
};

#endif // _DIMM_H
