/*
    $Id: indices.h 117 2010-03-19 12:19:16Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _INDICES_H
#define _INDICES_H

#include <stdexcept>
#include <list>
#include <vector>
#include <iterator>

#include <boost/bind.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/symmetric.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/tuple/tuple.hpp>

#include "table_storage.h"

namespace indices {
	using namespace boost::numeric::ublas;

	typedef boost::posix_time::ptime time_type;
	
	typedef double                value_type;
	typedef vector< value_type >  vector_type;
	typedef vector_type::size_type size_type;
	typedef symmetric_matrix< value_type > symmetric_matrix_type;
	typedef vector_type           flux_type;
	typedef symmetric_matrix_type vars_type;
	typedef symmetric_matrix_type rho1_type;
	typedef vector_type           rho2_type;
	typedef vector_type           indesis_type;
	typedef symmetric_matrix_type indices_type;

	flux_type flux_nonlinearity_correction( const flux_type& flux, const vector_type& deadtime );
	flux_type flux_background_correction( const flux_type& flux, const vector_type& scatter, const vector_type& background );
	boost::tuple< vars_type, rho1_type, rho2_type >
		vars_correction( const flux_type& flux, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2, const vector_type& deadtime, const vector_type& npoisson );
	boost::tuple< indices_type, indesis_type > calculate( const flux_type& flux, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 );

	struct indices_sample {
		time_type    timestamp;
		flux_type    flux;
		indices_type indices;
		indesis_type indesis;

		indices_sample( const time_type& time, const flux_type& flux, const indices_type& indices, const indesis_type& indesis ):
			timestamp(time), flux(flux), indesis(indesis), indices(indices) {}
	};

	const flux_type& flux( const indices_sample& x );
	const indices_type& indexes( const indices_sample& x );
	const indesis_type& indesis( const indices_sample& x );
};

class indices_storage:
	public table_storage< indices::indices_sample > {
public:
	typedef table_storage_column< indices::indices_sample, boost::function1< indices::flux_type, indices::indices_sample > >    flux_column_type;
	typedef table_storage_column< indices::indices_sample, boost::function1< indices::indices_type, indices::indices_sample > > indices_column_type;
	typedef table_storage_column< indices::indices_sample, boost::function1< indices::indesis_type, indices::indices_sample > > indesis_column_type;

	typedef boost::function2< bool, const indices::flux_type&, bool > filter_type;
	typedef std::list< filter_type > filter_list_type;
private:
	flux_column_type    m_flux_column;
	indices_column_type m_indices_column;
	indesis_column_type m_indesis_column;

	filter_list_type m_filter_list;
public:
	indices_storage( size_type reserve = 60 );

	void add_filter( const filter_type& filter ) {
		m_filter_list.push_back( filter );
	}
	
	const flux_column_type&    flux()    const { return m_flux_column; }
	const indices_column_type& indices() const { return m_indices_column; }
	const indesis_column_type& indesis() const { return m_indesis_column; }

	void push_back( const indices::time_type& time, const indices::flux_type& flux, const indices::indices_type& indices, const indices::indesis_type& indesis ) {
		table_storage< indices::indices_sample >::push_back( indices::indices_sample( time, flux, indices, indesis ) );
	}

	bool flux_filter( const indices::flux_type& flux, bool wexcept = true ) const;
};

class indices_threshold_filter {
	public:
		typedef indices::value_type value_type;
		typedef indices::flux_type  flux_type;
	private:
		value_type m_threshold;
	public:
		indices_threshold_filter( value_type threshold );
		~indices_threshold_filter();

		bool operator() ( const flux_type& flux, bool wexcept = true ) const;
};

#endif // _INDICES_H
