/*
    $Id: integrals.cpp 116 2010-03-19 10:46:46Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath>
#include <algorithm>
#include <numeric>
#include <boost/bind.hpp>
#include <boost/format.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/symmetric.hpp>
#include <lsp/singular_decomposition.h>

#include "typesdef.h"
#include "indices.h"
#include "integrals.h"

const double KM2M = 1000;
const double POWSEE = 0 ;       // Power of Cn2 moment for seeing
const double POWISP = 5.0/3.0;  // Power of Cn2 moment for isoplanatic angle
const double POWEFF = 1.0;      // Power of Cn2 moment for Effective altitude
const double POWISK = 2.0;      // Cn2 second moment power

const double MINRELW = 6e-4 ;   // Minimal ratio of the singular value to the maximal one

static const double powers[] = { POWSEE, POWISP, POWEFF, POWISK };

const unsigned int npower = sizeof(powers)/sizeof(powers[0]);

static const double HBOUND = 0.38;// Height of the boundary layer [kilometers] to fit the Cn2-integral
static const int TOPLAYERS =  4; // Number of rejected high-altitude layers
const double JPOW_SEE = 0.6 ;
const double LPOW_SEE = -0.2;
const double LEFF_SEE = 1.1486983549970351;
const double KS = 1.73588e+07;  // Calibration constant in seeing formula (AstL 45, p.399):
                                //\f$ seeing[''] = KS * \lambda_{eff}^{-1/5} M_0^{3/5} *
                                //\sec^{-1}(z)\f$, where \em lambda_eff is in [mkm]*/

const double JPOW_ISP = -0.6;   // Power of Cn-5/3th moment for isoplan. angle
const double LPOW_ISP = 1.2;    // Power of lambda for isoplan. angle
const double LEFF_ISP = 0.43527528164806206;
const double KP = 0.000756348;  // Calibration constant in isoplanatic angle formula (AstL 45, p.399)
                                // \f$ \theta_0[''] = KP * (M_{5/3} /\lambda_{eff}^2)^{-3/5} * \sec^{8/5}(z) \f$, where
                                // \f$ KP = (2.905*(2\pi)^2*(MKM2M)^{-2})^{-3/5}*206265 \f$*/
const double KT = 0.175;        // Atmospheric time constant empirical calibration constant for DESI for 1,3ms
                                // integrations in diameter=2cm apertures with -1km shift

const double SPOW_TC  = -0.6;   // Power of DESI for atm. time constant

template<class T> vector<T>& reject_elements( vector<T>& x, T minval ) {
    for( typename vector<T>::iterator it = x.begin(); it != x.end(); ++it ){
        *it = ( *it > minval )? *it : T(0);
    }
    return x;
}


integrals_factory::dcf_matrix_type integrals_factory::calculate( const std::string& spectral_type, const Weights& awf ) {
	typedef dcf_matrix_type            result_type;
	typedef dcf_matrix_type::size_type size_type;
	
	result_type ret( npower, awf.weight_functions_number()-2/*dimm*/ );

	size_type i = 0;
	vector<double> zgrid = awf.grid();
	while( zgrid( i ) < HBOUND ) i++;
	vector<double> h = subrange( zgrid, i, zgrid.size()-TOPLAYERS );
	
	Weights wfs( awf.apertures_number(), h );
	wfs.interpolate( awf ) ;

	for( unsigned int ipow = 0; ipow < npower ; ipow++ ) {
		double power = -powers[ipow]/2.0;
		double scale = pow( KM2M, powers[ipow] );
		vector<double> c = apply_to_all( apply_to_all( h, functor::log<double>() )*power, functor::exp<double>() );
		vector<double> b = apply_to_all( c, functor::inv<double>() )*scale;
		matrix<double> A( h.size(), ret.size2() );            // Make the matrices of weights
		for( unsigned int j = 0; j < ret.size2(); j++ ) {
			column( A, j ) = element_prod( row( wfs.weights(), j ), c );
		}
		matrix< double > left  = identity_matrix<double>( A.size1() );
		matrix< double > right = identity_matrix<double>( A.size2() );
		lsp::singular_decomposition< matrix< double > > sd( A );
		sd.apply( left, right );
		vector<double> s = diag_elements( A );
		reject_elements( s, A(0,0)*MINRELW );
		vector<double> x = element_prod( subrange( prod( left, b ), 0, s.size() ), apply_to_all( s, functor::inv<double>() ));
		row( ret, ipow ) = prod( right, x );
	}

	return ret;
}
integrals_factory::integrals_factory() {
}
integrals_factory::integrals_factory( const integrals_factory& factory ) {
	m_storage = factory.m_storage;
}
void integrals_factory::swap( integrals_factory& factory ) throw() {
	m_storage.swap( factory.m_storage );
}
integrals_factory& integrals_factory::operator=( const integrals_factory& factory ) {
	integrals_factory tmp( factory );
	tmp.swap( *this );
	return *this;
}
integrals_factory::~integrals_factory() {
}
integrals_factory::functor_type integrals_factory::get_functor( const std::string& spectral_type, const Weights& awf ) {
	storage_type::const_iterator it = m_storage.find( spectral_type );
	if( it != m_storage.end() )
		return functor_type( it->second.first, it->second.second );

	std::pair< storage_type::iterator, bool > ret = m_storage.insert( std::make_pair( spectral_type, std::make_pair( calculate( spectral_type, awf ), awf.lambda_eff() ) ) );

	return functor_type( ret.first->second.first, ret.first->second.second );
}

integrals_functor::result_type integrals_functor::operator() ( const indices_type& indices, const indesis_type& indesis, double mz ) const {
	assert( indices.size1() == 4 );
	assert( indices.size2() == 4 );
	assert( indesis.size() == 4 );

	vector< double > aints(10);
	aints(0) = indices(0,0); aints(1) = indices(1,1); aints(2) = indices(2,2); aints(3) = indices(3,3);
	aints(4) = indices(0,1); aints(5) = indices(0,2); aints(6) = indices(0,3);
	aints(7) = indices(1,2); aints(8) = indices(1,3); aints(9) = indices(2,3);

	aints = prod( m_dcf, aints );

	assert( npower <= aints );
	for( vector< double >::size_type i = 0; i < npower; ++i ){
		aints(i) = aints(i) / std::pow( mz, 1.0 + powers[i] );
	}
	
	result_type ret;

	ret.free_seeing       = KS * LEFF_SEE * std::pow( aints(0), JPOW_SEE );
	ret.isoplanatic_angle = KP * LEFF_ISP * std::pow( aints(1), JPOW_ISP );
	ret.altitude_eff      = 0.001 * std::pow( aints(1) / aints(2), 1.0/( POWISP-POWEFF ) );
	ret.time_constant     = KT * std::pow( indesis(0), SPOW_TC );
	ret.moment_0          = aints(0);
	ret.moment_2          = aints(3);

	return ret;
}

double integrals::free_seeing( const integrals_sample& x ) {
	return x.free_seeing;
}
double integrals::isoplanatic_angle( const integrals_sample& x ) {
	return x.isoplanatic_angle;
}
double integrals::altitude_eff( const integrals_sample& x ) {
	return x.altitude_eff;
}
double integrals::time_constant( const integrals_sample& x ) {
	return x.time_constant;
}
double integrals::moment_0( const integrals_sample& x ) {
	return x.moment_0;
}
double integrals::moment_2( const integrals_sample& x ) {
	return x.moment_2;
}

integrals_storage::integrals_storage( size_type reserve ):
	table_storage< integrals::integrals_sample >( reserve ),
	m_free_seeing_column( *this, boost::bind(&integrals::free_seeing,_1) ),
	m_isoplanatic_angle_column( *this, boost::bind(&integrals::isoplanatic_angle,_1) ),
	m_altitude_eff_column( *this, boost::bind(&integrals::altitude_eff,_1) ),
	m_time_constant_column( *this, boost::bind(&integrals::time_constant,_1) ),
	m_moment_0_column( *this, boost::bind(&integrals::moment_0,_1) ),
	m_moment_2_column( *this, boost::bind(&integrals::moment_2,_1) ) {
}
