/*
    $Id: integrals.h 115 2010-03-14 08:48:08Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _INTEGRALS_H
#define _INTEGRALS_H

#include <string>
#include <stdexcept>

#include <boost/tuple/tuple.hpp>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>

#include "table_storage.h"
#include "nocase_less.h"

#include "typesdef.h"
#include "weif.h"

namespace integrals {
	
	struct integrals_sample {
		double free_seeing;
		double isoplanatic_angle;
		double altitude_eff;
		double time_constant;
		double moment_0;
		double moment_2;
	};

	double free_seeing( const integrals_sample& x );
	double isoplanatic_angle( const integrals_sample& x );
	double altitude_eff( const integrals_sample& x );
	double time_constant( const integrals_sample& x );
	double moment_0( const integrals_sample& x );
	double moment_2( const integrals_sample& x );
};
using namespace boost::numeric::ublas;

class integrals_factory;
class integrals_functor;

class integrals_factory {
	public:
		typedef matrix< double > dcf_matrix_type;
		typedef std::map< std::string, std::pair< dcf_matrix_type, double >, nocase_less< std::string > > storage_type;
		typedef integrals_functor functor_type;
		typedef storage_type::const_iterator const_iterator;
	private:
		storage_type m_storage;

		dcf_matrix_type calculate( const std::string& spectral_type, const Weights& awf );
	public:
		integrals_factory();
		integrals_factory( const integrals_factory& factory );
		void swap( integrals_factory& factory ) throw();
		integrals_factory& operator=( const integrals_factory& factory );
		~integrals_factory();

		functor_type get_functor( const std::string& spectral_type, const Weights& awf );
};

class integrals_functor {
	public:
		typedef integrals_factory             factory_type;
		typedef factory_type::dcf_matrix_type dcf_matrix_type;

		typedef integrals::integrals_sample result_type;

		typedef symmetric_matrix<double> indices_type;
		typedef vector<double> indesis_type;
	private:
		const dcf_matrix_type& m_dcf;
		double m_lambda_eff;
	public:
		integrals_functor( const dcf_matrix_type& dcf, double lambda_eff ):
			m_dcf( dcf ), m_lambda_eff( lambda_eff ) {}

		result_type operator() ( const indices_type& indices, const indesis_type& indesis, double mz ) const;
};

class integrals_storage:
	public table_storage< integrals::integrals_sample > {
public:
	typedef table_storage_column< integrals::integrals_sample, boost::function1< double, integrals::integrals_sample > > common_column_type;
	typedef common_column_type free_seeing_column_type;
	typedef common_column_type isoplanatic_angle_column_type;
	typedef common_column_type altitude_eff_column_type;
	typedef common_column_type time_constant_column_type;
	typedef common_column_type moment_0_column_type;
	typedef common_column_type moment_2_column_type;
private:
	free_seeing_column_type       m_free_seeing_column;
	isoplanatic_angle_column_type m_isoplanatic_angle_column;
	altitude_eff_column_type      m_altitude_eff_column;
	time_constant_column_type     m_time_constant_column;
	moment_0_column_type          m_moment_0_column;
	moment_2_column_type          m_moment_2_column;
public:
	integrals_storage( size_type reserve = 60 );

	const free_seeing_column_type&       free_seeing()       const { return m_free_seeing_column; }
	const isoplanatic_angle_column_type& isoplanatic_angle() const { return m_isoplanatic_angle_column; }
	const altitude_eff_column_type&      altitude_eff()      const { return m_altitude_eff_column; }
	const time_constant_column_type&     time_constant()     const { return m_time_constant_column; }
	const moment_0_column_type&          moment_0()          const { return m_moment_0_column; }
	const moment_2_column_type&          moment_2()          const { return m_moment_2_column; }

	void push_back( const integrals::integrals_sample& integrals ) {
		table_storage< integrals::integrals_sample >::push_back( integrals );
	}
};


#endif // _INTEGRALS_H
