/*
    $Id: nr.cpp 101 2010-03-02 08:36:15Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath>
#include "nr.h"

/**< nr.h is modified! erf*() are commented to avoid conflicts with math.h;
float->double; also all declarations of utilities are included there by:
#include "nrutil.h" in NR_CPP mode. The utilities are converted, to obey the
rules of C++
- declarations of external functions are removed from the function bodies
(instead, #include "nr.h" serves);
- arguments are supplied with their types in parentheses (nothing is thus in
between the arguments in parentheses and the function body {...},
- all the function names are prefixed with nr:: (i.e. belong to namespace "nr");
- all the calls to nrerror() are followed by instruction 'return'. */

void nr::four1( ublas::vector<double>& data  ) {
    int n,mmax,m,j,istep,i;
    double wtemp,wr,wpr,wpi,wi,theta;
    double tempr,tempi;

    n = data.size();
    j = 1;
    for( i = 1; i < n; i += 2 ) {
        if( j > i ) {
            std::swap( data(j-1), data(i-1) );
            std::swap( data(j),data(i) );
        }
        m = n/2;
        while( m >= 2 && j > m ) {
            j -= m;
            m /= 2;
        }
        j += m;
    }
    mmax = 2;
    while( n > mmax ) {
        istep = 2*mmax;
        theta = 2*M_PI/mmax;
        wtemp = sin( 0.5*theta );
        wpr = -2.0*wtemp*wtemp;
        wpi = sin( theta );
        wr = 1.0;
        wi = 0.0;
        for( m = 1; m < mmax; m += 2 ) {
            for( i = m; i <= n; i += istep ) {
                j = i + mmax;
                tempr = wr*data[j-1] - wi*data[j];
                tempi = wr*data[j] + wi*data[j-1];
                data[j-1] = data[i-1] - tempr;
                data[j] = data[i] - tempi;
                data[i-1] += tempr;
                data[i] += tempi;
            }
            wtemp = wr;
            wr = wr*wpr - wi*wpi + wr;
            wi = wi*wpr + wtemp*wpi + wi;
        }
        mmax = istep;
    }
}

void nr::spline( const ublas::vector<double>& x, const ublas::vector<double>& y, double yp1, double ypn, ublas::vector<double>& y2 ) {
    assert(  x.size() ==  y.size() );
    double p, qn, sig, un;
    int n = x.size();
    ublas::vector<double> u( n );
    if( yp1 > 0.99e30 ) y2(0) = u(0) = 0.0;
    else {
        y2(0) = -0.5;
        u(0) = ( 3.0/(x(1)-x(0)) )*( (y(1)-y(0) )/( x(1)-x(0) ) - yp1 );
    }
    for( int i = 1; i < n-1; i++ ) {
        sig = ( x(i)-x(i-1) )/( x(i+1)-x(i-1) );
        p = sig*y2(i-1) + 2.0;
        y2(i) = ( sig - 1.0 )/p;
        u(i) = ( y(i+1) - y(i) )/( x(i+1)-x(i) ) - ( y(i) - y(i-1) )/( x(i)-x(i-1) );
        u(i) = ( 6.0*u(i)/( x(i+1)-x(i-1) ) - sig*u(i-1) )/p;
    }
    if( ypn > 0.99e30 ) qn = un = 0.0;
    else {
        qn = 0.5;
        un = ( 3.0/( x(n-1)-x(n-2) ) )*( ypn - ( y(n-1) - y(n-2) )/( x(n-1) - x(n-2) ));
    }
    y2(n-1) = ( un - qn*u(n-2) )/( qn*y2(n-2) + 1.0 );
    for( int k = n-2; k >= 0; k-- ) y2(k) = y2(k)*y2(k+1) + u(k);
}

void nr::splint( const ublas::vector<double>& xa, const ublas::vector<double>& ya, ublas::vector<double>& y2a, double x, double& y ) {
    int klo = 0;
    int khi = xa.size()-1;
    int k = 0;
    while( khi - klo > 1 ) {
        k = ( khi + klo )/2;
        if( xa(k) > x ) khi = k;
        else klo = k;
    }
    double h = xa(khi) - xa(klo);
    assert( h != 0 );
    double a = ( xa(khi) - x )/h;
    double b = ( x - xa(klo) )/h;
    y = a*ya(klo) + b*ya(khi) + (( a*a*a - a )*y2a(klo) + ( b*b*b - b )*y2a(khi) )*h*h/6.0;
}
