/*
    $Id: profile.cpp 115 2010-03-14 08:48:08Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <lsp/nnls.h>
#include <boost/bind.hpp>
#include <boost/format.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/triangular.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <iostream>
#include <iomanip>
#include <algorithm>
#include <numeric>

#include "io.h"
#include "profile.h"
#include "typesdef.h"

const double MAXCHI2 = 999.0;
const double JPOW_SEE = 0.6 ;
const double LPOW_SEE = -0.2;
const double KS = 1.73588e+07 ;

using namespace boost::numeric::ublas;

restoration_functor::restoration_functor( const vector<double>& grid, const symmetric_matrix< double >& corr_matrix, double secz, const Weights& basewf ):
	m_grid( grid ), m_cmatrix(corr_matrix), m_secz(secz), m_lambda_eff( basewf.lambda_eff() ) {
		
	if( zero_elements( diag_elements( m_cmatrix ) ) > 0 ) {
		throw std::logic_error( (boost::format("Zero variances: %1% x %2% %3%") % m_cmatrix.size1() % m_cmatrix.size2()  % vector<double>( diag_elements( m_cmatrix )) ).str().c_str() );
	}

	lsp::cholesky_decomposition( m_cmatrix, upper_tag() );
	triangular_adaptor< symmetric_matrix< double >, lower > R( m_cmatrix );
	m_wmatrix = identity_matrix< double >( corr_matrix.size1() );
	inplace_solve( R, m_wmatrix, lower_tag() );
	
	if( zero_elements( diag_elements( m_wmatrix ) ) > 0 ) {
		throw std::logic_error( (boost::format("Zero weights: %1%") % vector<double>( diag_elements( m_wmatrix )) ).str().c_str() );
	}
		
	Weights wfs( basewf.apertures_number(), grid*secz );
	wfs.interpolate( basewf );
	m_pmatrix = prod( m_wmatrix, subrange(wfs.weights(),0,m_wmatrix.size2(),0,wfs.weights().size2()) );
}
vector< double > restoration_functor::flat_reorder( const matrix<double>& indices ) {
	vector< double > inds(10);
	inds(0) = indices(0,0); inds(1) = indices(1,1); inds(2) = indices(2,2); inds(3) = indices(3,3);
	inds(4) = indices(0,1); inds(5) = indices(0,2); inds(6) = indices(0,3);
	inds(7) = indices(1,2); inds(8) = indices(1,3); inds(9) = indices(2,3);
	return inds;
}
restoration_functor::result_type restoration_functor::operator() ( const vector<double>& indices ) const {
	vector< double > inds = prod( m_wmatrix, indices );

	result_type ret;
	
	matrix< double > cov = zero_matrix<double>( m_pmatrix.size2(), m_pmatrix.size2() );

	lsp::nnls< matrix< double >, vector< double > > nnls( m_pmatrix, inds );
	nnls.solve( ret.cn2, cov );
	
	ret.residual = prod( m_pmatrix, ret.cn2 ) - inds;

	ret.chi2 = std::min( inner_prod( ret.residual, ret.residual ), MAXCHI2 );

	ret.cn2 /= m_secz;
	return ret;
}
double restoration_functor::seeing( const vector<double>& cn2 ) const {
	double cn2_sum = sum( cn2 );
	double secz = 1.0; // already corrected cn2 at input
	return KS*pow( cn2_sum/secz, JPOW_SEE )*pow( m_lambda_eff, LPOW_SEE );
}
vector<double> restoration_functor::indices_residual( const vector<double>& residuals ) const {
	return prod( triangular_adaptor< const matrix< double >, lower >(m_cmatrix), residuals );
}

profile::chi2_type profile::chi2( const profile_sample& x ) {
	return x.chi2;
}
const profile::cn2_type& profile::cn2( const profile_sample& x ) {
	return x.cn2;
}
const profile::residual_type& profile::residual( const profile_sample& x ) {
	return x.residual;
}

profile_storage::profile_storage( size_type reserve ):
	table_storage< profile::profile_sample >( reserve ),
	m_cn2_column( *this, boost::bind(&profile::cn2, _1) ),
	m_residual_column( *this, boost::bind(&profile::residual, _1)  ),
	m_chi2_column( *this, boost::bind(&profile::chi2, _1) )
{
}
