/*
    $Id: profile.h 115 2010-03-14 08:48:08Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _PROFILE_H
#define _PROFILE_H

#include <string>
#include <stdexcept>

#include <lsp/cholesky_decomposition.h>

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/symmetric.hpp>

#include "table_storage.h"
#include "weif.h"

namespace profile {
	using namespace boost::numeric::ublas;

	typedef double                 value_type;
	typedef vector<value_type>     vector_type;
	typedef vector_type::size_type size_type;

	typedef vector_type cn2_type;
	typedef vector_type residual_type;
	typedef value_type  chi2_type;

	struct profile_sample {
		cn2_type       cn2;
		residual_type  residual;
		chi2_type      chi2;
	};

	chi2_type chi2( const profile_sample& x );
	const cn2_type& cn2( const profile_sample& x );
	const residual_type& residual( const profile_sample& x );
};

template<class VE> typename VE::size_type zero_elements( const vector_expression<VE>&x ) {
    typename VE::size_type n = 0;
    for( typename VE::size_type j = 0; j < x().size(); j++ )
	    n += ( x()(j) == 0 || std::isinf(x()(j)) || std::isnan(x()(j)) ) ? 1 : 0;
    return n;
}

template<class ME> vector< typename ME::value_type > diag_elements( const matrix_expression<ME> &x ) {
    typename ME::size_type r = std::min( x().size1(), x().size2() );
    slice s(0, 1, r );
    return matrix_vector_slice< const ME > ( x, s, s );
}

class restoration_functor {
	public:
		typedef profile::profile_sample result_type;

		typedef vector<double> grid_type;
		typedef matrix<double> weight_matrix;
	private:
		grid_type     m_grid;
		weight_matrix m_wmatrix;
		symmetric_matrix< double > m_cmatrix;
		matrix< double > m_pmatrix;
		double        m_secz;
		double        m_lambda_eff;

	public:
		static vector< double > flat_reorder( const matrix<double>& indices );
	
		restoration_functor( const vector<double>& grid, const symmetric_matrix< double >& corr_matrix, double secz, const Weights& basewf );

		result_type operator() ( const vector<double>& indices ) const;

		double seeing( const vector<double>& cn2 ) const;
		vector<double> indices_residual( const vector<double>& residuals ) const;
};

class profile_storage:
	public table_storage< profile::profile_sample > {
public:
	typedef table_storage_column< profile::profile_sample, boost::function1< profile::cn2_type,      profile::profile_sample > > cn2_column_type;
	typedef table_storage_column< profile::profile_sample, boost::function1< profile::residual_type, profile::profile_sample > > residual_column_type;
	typedef table_storage_column< profile::profile_sample, boost::function1< profile::chi2_type,     profile::profile_sample > > chi2_column_type;
private:
	cn2_column_type      m_cn2_column;
	residual_column_type m_residual_column;
	chi2_column_type     m_chi2_column;
public:
	profile_storage( size_type reserve = 60 );

	const cn2_column_type&      cn2()      const { return m_cn2_column; }
	const residual_column_type& residual() const { return m_residual_column; }
	const chi2_column_type&     chi2()     const { return m_chi2_column; }
	
	void push_back( const profile::profile_sample& profile ) {
		table_storage< profile::profile_sample >::push_back( profile );
	}
};

#endif // _PROFILE_H
