/*
    $Id: weif.cpp 102 2010-03-02 18:04:35Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/****************************************************************************/
/** @file weif.cpp
                 MASS project: TURBINA module file
                             weif.cpp
                    Description is given in weif.hxx                        */
/* Editor tabular size: 4                                                   */
/****************************************************************************/

/* Insert the definition WEIFTEST in compilation for the module self-tests:
gcc -c weif.c .... -DWEIFTEST=1 */

#include <fstream>
#include <iostream>
#include <iomanip>
#include <vector>
#include <algorithm>
#include <functional>
#include <numeric>
#include <cmath>
#include <limits>

#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>

#include <gsl/gsl_integration.h>

#include "io.h"
#include "typesdef.h"
#include "weif.h"
#include "nr.h"

#ifdef WEIFTEST

/*------------------------------ TESTING PART --------------------------*/
const int max_nodes = 20;               // number of apertures in segmentor :
const int max_naper = 10;               // number of apertures in segmentor :
                            // Effective size of segmentor annular apertures [cm] :
double cf_aper_inn[max_naper] = { 0.000, 1.648, 3.061,  5.730 };
double cf_aper_out[max_naper] = { 1.617, 2.983, 5.652, 10.595};
                                            // Magnification of the system (multiplier of cf_aper_* sizes):
double cf_magnif = 1.0;
/** Usual value for low boundary of altitude range */
const double Z0 = 0 ;

template<class T, int nmax > int str2da( T* a, const std::string& cs ) {
    std::string s(cs);
    std::string::size_type i = 0;
    while( (i = s.find_first_of( ",", i )) != std::string::npos ) s.at(i) = ' ';
    int n = 0;
    std::istringstream istr(s);
    while( n < nmax && istr >> a[n] ) n++;
    return n;
}

int main( int argc, char **argv ) {
    int cf_naper = 4;
    int nodes = 0;

    std::string rfile = "/opt/turbina/data/data/mass.crv";
    std::string sfile;
    std::string cfile ="ccd.crv";
    std::string wfile = "/dev/stdout";
    double d[max_naper];
    double zgrid[max_nodes];
	int c;                                  // parse options:
	while( ( c = getopt(argc,argv,"r:c:s:N:Z:I:O:hv") ) != -1  ) {
		if( argc < 3 ) c = 'h';             // no spectra and sed file(s) - output help
		int nap = max_naper;
		switch (c) {
			case 'r':    rfile = std::string(optarg); break;
			case 's':    sfile = std::string(optarg); break;
			case 'c':    cfile = std::string(optarg); break;
			case 'N':  { int n=atoi(optarg); if ( n>0 && n<=max_naper ) cf_naper = n; break; }
			case 'I':    nap = str2da<double,max_naper>( cf_aper_inn, std::string(optarg) ); break;
			case 'O':    nap = str2da<double,max_naper>( cf_aper_out, std::string(optarg) ); break;
			case 'Z':    nodes = str2da<double,max_nodes>( zgrid, std::string(optarg) );  break;
			case 'h':
            default :
				std::cerr << "WEIF: scintillation weight functions computer" << std::endl;
				"Usage: \n./weif -ssfile [-rrfile] [-Nn] [-Ii1,..in -Oo1,..on] [-Zz1,z2,z3,z4] >out.wf\n"
				"  Input data files:\n"
				"\t-s spectrum-file \n"
				"\t-r response-file (default: /opt/turbina/data/data/mass.crv)\n"
				"  Options:\n"
				"\t-N number of apertures (default: 4)\n"
				"\t-I innerDiam[0],...,innerDiam[N-1] (in cm, default: old MASS @k=15.7)\n"
				"\t-O outerDiam[0],...,outerDiam[N-1] (in cm, default: old MASS @k=15.7)\n"
				"\t-Z zmin,zmax,zstep,zminstep - altitude grid parameters, km \n\t\t(default: 0,40,0.2,0.04)\n";
				exit (-1);
		}
	}
    Weights cf_wf( cf_naper, wf::grid_generate( nodes, zgrid ) );
    vector<double> cf_eps_inner(cf_naper);
    vector<double> cf_eps_outer(cf_naper);
    for( int i = 0; i < cf_naper; i++ ) cf_eps_inner(i) = cf_aper_inn[i]*cf_magnif;
    for( int i = 0; i < cf_naper; i++ ) cf_eps_outer(i) = cf_aper_out[i]*cf_magnif;
    Spectral_data spec( rfile, wfile, sfile, cfile );
	cf_wf.calculation( spec, cf_eps_inner, cf_eps_outer, 9, 19.6 );
    return 0;
}

#endif                    // ifdef WEIFTEST end
/*----------------------- TESTING PART END --------------------------------*/

template <class T> T nexthigher(T k) {
    if (k == 0) return 1;
    k--;
    for (int i=1; i < std::numeric_limits<T>::digits; i<<=1) k = k | k >> i;
    return k+1;
}

const double POWIND = -2.66666666667;
const double WSCALE = 1E11 ;
const double MKM2CM = 1E-4 ;                // Conversion of [mkm] to [cm]:
const double DER0  = 0.0 ;                  // Splines: define the first derivative of weights at z=0 equal to 0:
const double DERNAT  = 1.0E31 ;             // signal to set the "natural" upper boundary cond.: 2nd weight der.=0

Weights::Weights( int n, const vector<double>& z ) : mass_nchan(n), weights_number(n*(n+1)/2+2), zgrid(z) {
    ws.resize( weights_number, zgrid.size() );
}

void Weights::write( const std::string& filename ) const {
    static std::string name[] = { "A", "B", "C", "D", "AB", "AC", "AD", "BC", "BD", "CD", "L", "T" };
    assert( ws.size1() != 0 && ws.size2() != 0 );
    std::fstream file;
    file.open( filename.c_str(), std::ios::out );
    if( !file ) throw std::runtime_error( "File " + filename + " open error" );
    if( !(file << "# N " << eps_inn.size() << std::endl) ) throw std::runtime_error( "File " + filename + " write error" );;
    for( int i = 0; i < eps_inn.size(); i++ ) {
        file << "# " << name[i] << " " << std::fixed << std::setprecision(2) << eps_inn(i) << " " << eps_out(i) << std::endl;
    }
    file << "# S " << std::fixed << std::setprecision(4) << leff << " " << lblue << " " << lred << std::endl;
    file << "# Z " << std::setprecision(2) << zgrid << std::endl;
    file << "# W " << std::setprecision(0) << std::scientific << WSCALE << std::endl;
    file << "##Altit.";
    for( int i = 0; i < ws.size1(); i++ ) file << " W(" << name[i] << ")/" <<  std::scientific << WSCALE;
    file << std::endl;
	for( int i = 0; i < zgrid.size(); i++) {
        file << std::fixed << std::setprecision(3) <<  std::setw(7) <<  zgrid(i) << " ";
		for( int j = 0; j < ws.size1(); j++) {
            file << " " << std::scientific << std::setprecision(4) <<  std::setw(10) << ws(j,i)/WSCALE;
		}
        file << std::endl;
    }
}

template<class T>  bool more_dense( const vector<T>& x, const vector<T>& y ) {
    if( x.size() == 0 ) return false;
    if( y.size() == 0 ) return true;
    return x(0) < y(0) && x( x.size()-1 ) > y( y.size()-1 ) && x.size() > y.size();
}

void Weights::interpolate( const Weights& refwf ) {
           // Check entries and parameters: new_wf zgrid should be enclosed in z-range of the old wf grid!
    assert( refwf.ws.size1() == refwf.weights_number && refwf.ws.size2() == refwf.zgrid.size() );
    assert( zgrid.size() > 0 );
    if( more_dense( zgrid, refwf.zgrid ) ) throw std::runtime_error( "Attempt to interpolate from narrower zgrid" );
    ws.resize( refwf.ws.size1(), zgrid.size() );
    for( int w = 0; w < refwf.ws.size1(); w++ ) {
        Spline base( refwf.zgrid, row( refwf.ws, w ), (( refwf.ws(w,0) != 0 ) ? DERNAT : DER0 ), DERNAT );
        for( int iz = 0; iz < zgrid.size(); iz++) ws(w,iz) = base.get( zgrid(iz) );
    }
    leff = refwf.leff;
    lblue = refwf.lblue;
    lred = refwf.lred;
    eps_inn = refwf.eps_inn;
    eps_out = refwf.eps_out;
}
 
int Weights::calcleff( vector<double>& lambda, vector<double>& edist ) {
    assert( edist.size() == lambda.size() );
    leff = inner_prod( lambda, edist )/ sum( edist );
    if( isinf(leff) || isnan(leff) ) throw std::runtime_error( "Illegal response data" );
    vector<double> cumsum( edist.size() );
    partial_sum( edist.begin(), edist.end(), cumsum.begin() );
    cumsum /= cumsum(edist.size()-1);
    vector<double>::iterator pos25 = std::lower_bound( cumsum.begin(), cumsum.end(), 0.25 );
    int b = pos25.index();
    lblue = lambda(b-1) + ( 0.25 - cumsum(b-1) )*( lambda(b) - lambda(b-1) )/( cumsum(b) - cumsum(b-1) );
    vector<double>::iterator pos75 = std::lower_bound( cumsum.begin(), cumsum.end(), 0.75 );
    int r = pos75.index();
    lred  = lambda(r-1) + ( 0.75 - cumsum(r-1) )*( lambda(r) - lambda(r-1) )/( cumsum(r) - cumsum(r-1) );
    vector<double>::iterator eff_pos = std::lower_bound( lambda.begin(), lambda.end(), leff );
    return eff_pos.index();
}

const int MKM_RANGE  = 0 ;   // Units of optical range wavelength: 0 for [mkm], 3 for [nm], 4 for [\AA]

void Weights::read_crv( const std::string& filename, vector<double>& col1, vector<double>& col2  ) {
    std::ifstream file( filename.c_str() );
    if( file.fail() ) throw std::runtime_error( "File " + filename + " open error" );
    std::vector<double> x;
    std::vector<double> y;
    std::string iline;
    while( getline( file, iline ) ) {
        if( iline.size() < 1 || iline.at(0) == '#' ) continue;
        std::istringstream istr( iline );
        double cx, cy;
        if( !(istr >> cx >> cy) )  throw std::runtime_error( "File " + filename + " line format error" );
        x.push_back( cx );
        y.push_back( cy );
    }
    if( x.size() == 0 ) throw std::runtime_error( "File " + filename + " contains no data" );
    if( std::adjacent_find( x.begin(), x.end(), std::greater<double>() ) != x.end() ) throw std::runtime_error( "File " + filename + " contains unsorted data" );
    if( std::accumulate( y.begin(), y.end(), 0.0 ) <= 0.0 ) std::runtime_error( "File " + filename + " contains zero or negative data" );
    col1.resize( x.size() );
    col2.resize( x.size() );
    std::copy( x.begin(), x.end(), col1.begin() );
                            //Convert array of optical (!) wavelengths into given scale
    double conv = pow( 10.0, MKM_RANGE - ((int)log10( col1(col1.size()/2))+1) );
    col1 *= conv;
    std::copy( y.begin(), y.end(), col2.begin() );
}

void Weights::photon_distr( vector<double>& slambda, vector<double>& sed, vector<double>& rlambda, vector<double>& response ) {
                            // Find the starting point of the response curve at SED wavelength grid:
    vector<double>::iterator pos;
    pos = std::find( slambda.begin(), slambda.end(), rlambda(0) );
    if( pos ==  slambda.end() || !std::equal( pos, pos+rlambda.size(), rlambda.begin() ) )
                throw std::runtime_error( "No match of response and spectrum grids" );
    vector<double> resed = element_prod( subrange( sed, pos.index(), pos.index()+response.size() ), response );
    if( std::count_if( resed.begin(), resed.end(), std::bind2nd(std::less<double>(), 0.0 ) ) > 0 )
        throw std::runtime_error( "WEIF: SED has negative elements" );
    double enorm = sum( resed );       // Normalize SED:
    if( enorm <= 0 ) throw std::runtime_error( "WEIF: SED has non-positive norm" );
    sed = resed/enorm;
}

void Weights::calculation( const Spectral_data& specfile, const vector<double>& ap_inn, const vector<double>& ap_out, double dimmsize, double dimmbase ) {
    vector<double> rlambda;
    vector<double> response;
    vector<double> dlambda;
    vector<double> ccdcrv;
    vector<double> slambda;
    vector<double> mass_sed;
    vector<double> dimm_sed;
    assert( ap_inn.size() == mass_nchan && ap_out.size() == mass_nchan );
    read_crv( specfile.response, rlambda, response );   // Read the response function from the file:
    read_crv( specfile.ccdresponse, dlambda, ccdcrv );
    read_crv( specfile.spectra, slambda, mass_sed );
    dimm_sed = mass_sed;
    photon_distr( slambda, mass_sed, rlambda, response );    // Check that response curve wavelength grid complies the spectrum and multiply them
    photon_distr( slambda, dimm_sed, dlambda, ccdcrv   );    // Check that response curve wavelength grid complies the spectrum and multiply them
    eps_inn = ap_inn;
    eps_out = ap_out;
    dimm_aperture = dimmsize * 100;
    dimm_base = dimmbase * 100;
    mass_sed = element_div( mass_sed, rlambda*MKM2CM );
    calc( rlambda, mass_sed, dlambda, dimm_sed );
}

void Weights::calc( vector<double>& lambda, vector<double>& mass_dist, vector<double>& ccdlam, vector<double>& dimm_dist ) {
    assert( zgrid.size() >  0 );
    Spline* re_sfun;
    Spline* im_sfun;
    int shift_index = calcleff( lambda, mass_dist );

    spectral_band_fft( lambda, mass_dist, shift_index, re_sfun, im_sfun );
    int nnw = 0;                                // Counter of weighting functions
    for (int a = 0; a < mass_nchan; a++) {      // Normal weighting functions:
       row( ws, nnw++ ) = mass_weight( zgrid, eps_out(a), eps_inn(a), 0, 0, lambda(shift_index), re_sfun, im_sfun );
    } 
    for (int a = 0; a < mass_nchan - 1; a++) {  // Differential weighting functions:
        for (int b = a + 1; b < mass_nchan; b++) {
            row( ws, nnw++ ) = mass_weight( zgrid, eps_out(a), eps_inn(a), eps_out(b), eps_inn(b), lambda(shift_index), re_sfun, im_sfun );
        }
    }
    delete re_sfun;
    delete im_sfun;

    int dimm_index = calcleff( ccdlam, dimm_dist );
    spectral_band_fft( ccdlam, dimm_dist, dimm_index, re_sfun, im_sfun );
    row( ws, nnw++ ) = dimm_weight( zgrid, dimm_aperture, dimm_base, 0.0, ccdlam(dimm_index) , re_sfun, im_sfun ) * 1e+11;
    row( ws, nnw++ ) = dimm_weight( zgrid, dimm_aperture, dimm_base, M_PI/2, ccdlam(dimm_index) , re_sfun, im_sfun ) * 1e+11;
    delete re_sfun;
    delete im_sfun;
}

/*==================== weight calculation section =====================*/

void Weights::spectral_band_fft( vector<double>& lambda, vector<double>& edist, int center, Spline*& re_sfun, Spline*& im_sfun ) {
    const int GRIDFACT = 4 ;    // 0 - no stretch, 1 - stretch twice, 2 - stretch 4-times etc.
    double dlambda = lambda(1) - lambda(0);
    int edist_ftnn = 2*GRIDFACT*nexthigher(edist.size());
    vector<double> edist_ft = zero_vector< double >( edist_ftnn*2 );
	vector<double> edist_freq( edist_ftnn/2 );
    for( int l = 0; l < edist_freq.size(); l++ ) edist_freq(l) = l/(dlambda*MKM2CM)/edist_ftnn;
    for( int l = 0; l < edist.size(); l++ ) {      // Fill in the data array for FFT:
        int j = ( l-center < 0 )? edist_ftnn + l - center : l - center;
        edist_ft[2*j] = edist[l];
    }
    nr::four1( edist_ft );      // Compute the Fourier transform of the center-shifted SED:
    vector<double> re_edist_ft = vector_slice< vector<double> >( edist_ft, slice( 0, 2, edist_ftnn/2 ) );
    vector<double> im_edist_ft = vector_slice< vector<double> >( edist_ft, slice( 1, 2, edist_ftnn/2 ) );
    re_sfun = new Spline( edist_freq, re_edist_ft, DERNAT, DERNAT );
    im_sfun = new Spline( edist_freq, im_edist_ft, DERNAT, DERNAT );
}

double massf(double x, void * params) {
    mass_params p = *( mass_params *) params;
    double apert1 = 0.0, apert2 = 0.0, apert3 = 0.0, apert4 = 0;
    double f = p.altitude*x*x/2;
    double sf = p.im_sfun->get(f)*cos( p.PIx2lambda0*f ) + p.re_sfun->get(f)*sin( p.PIx2lambda0*f );
    sf = sf*sf*pow( x, POWIND );
    if( p.diameter1 != p.diameter2 ) {                // First aperture:
        apert1 = p.diameter1*j1( p.diameter1*x );
        apert2 = p.diameter2*j1( p.diameter2*x );
        apert1 = ( apert1 - apert2 )/( p.diameter1*p.diameter1 - p.diameter2*p.diameter2 );
    }
    if( p.diameter3 != p.diameter4) {                 // Second aperture: the same with indices 1->3, 2->4:
        apert3 = p.diameter3*j1( p.diameter3*x );
        apert4 = ( p.diameter4 == 0 )? 0 : p.diameter4*j1( p.diameter4*x );
        apert3 = ( apert3 - apert4 )/( p.diameter3*p.diameter3 - p.diameter4*p.diameter4 );
    }
    apert1 = 2.0*( apert1 - apert3 )/x;
    return sf*apert1*apert1;
}

const double POWDIFF = -0.666666666667;

double dimmf(double x, void * params) {         // Z-tilt case
    dimm_params p = *( dimm_params *) params;
    double f = p.altitude*x*x/2;
    double a = p.aperture*x;
    double b = p.dimmbase*x;
    double cf = p.re_sfun->get(f)*cos( p.PIx2lambda0*f ) - p.im_sfun->get(f)*sin( p.PIx2lambda0*f );
    cf = cf*8*jn(2,a)/a/a;
    cf = cf*cf*pow( x, POWDIFF );
    return cf*( 1.0 - 2.0*cos(p.phi)*cos(p.phi)*j0(b) + 2.0*cos(2*p.phi)*j1(b)/b );
}

const double EPS = 1.0e-4 ;     // Relative precision for integration
                                // CALIBR = (32*PI^3*9.69*1E-3)*(0.01^(5/3))*1E4;
                                // where 0.01^(5/3) comes from integration in [cm] instead of [m];
                                // 1E4 originates from conversion of [1/cm]^2 into [1/m]^2 in integration of sfunc()
const double CALIBR  = 44.62621363;
const double KM2CM  = 1e5 ;     // Conversion of [km] into [cm]:
const int  WORKSPACE = 10000;

vector<double> Weights::mass_weight( const vector<double>& z, double eps1, double eps2, double eps3, double eps4, double lambda0, Spline* re_sfun, Spline* im_sfun ) {

    gsl_integration_workspace * w = gsl_integration_workspace_alloc(WORKSPACE);
    double error;
    mass_params params;
    gsl_function F;
    F.function = &massf;
    F.params = &params;
    params.diameter1 = M_PI*eps1;
    params.diameter2 = M_PI*eps2;
    params.diameter3 = M_PI*eps3;
    params.diameter4 = M_PI*eps4;
    params.PIx2lambda0 = 2*M_PI*lambda0*MKM2CM;
    params.re_sfun = re_sfun;
    params.im_sfun = im_sfun;
    vector<double> result( z.size() );
    for( unsigned int i = 0; i < z.size(); i++ ) {
        params.altitude  = z(i)*KM2CM;
        gsl_integration_qagiu( &F, 0, 0, EPS, WORKSPACE, w, &result(i), &error);
    }
    gsl_integration_workspace_free( w );
    return result*CALIBR;
}

linear_generator::linear_generator( result_type start, result_type step ): start_(start), step_(step) {
}
linear_generator::result_type linear_generator::operator()( size_type i ) const {
	return ( start_ + step_ * i );
}
linear_generator::size_type linear_generator::size( result_type max ) const {
	return static_cast<size_type>( std::floor( (max - start_) / step_ ) );
}

lg_generator::lg_generator( result_type start, result_type lstep, result_type gstep ): start_(start), t_(lstep), q_(result_type(1) + gstep / 2) {
	max_ = static_cast< size_type >( std::floor( result_type(2) / gstep - start / lstep ) );
}
lg_generator::result_type lg_generator::operator()( size_type i ) const {
	if( i <= max_ ) return ( start_ + t_ * i );
	return ( start_ + t_ * max_ ) * std::pow( q_, static_cast<int>(i - max_) ); // FIXME: pow(double,int) for <4.2.1
}
lg_generator::size_type lg_generator::size( result_type max ) const {
	result_type gmax = std::floor( std::log( max / ( start_ + t_ * max_ ) ) / std::log( q_ ) );
	if( gmax > result_type(0) ) return static_cast< size_type >( gmax ) + max_ + 1;
	return static_cast< size_type >(std::floor( (max - start_) / t_ ));
}

const double DIMMCONST  = 2.403*4.64159;  // 4.64159 - conversion from cm to m

vector<double> Weights::dimm_weight( const vector<double>& z, double diam, double dbase, double phi, double lambda0, Spline* re_sfun, Spline* im_sfun ) {

    gsl_integration_workspace * w = gsl_integration_workspace_alloc(WORKSPACE);
    double error;
    dimm_params params;
    gsl_function F;
    F.function = &dimmf;
    F.params = &params;
    params.aperture = M_PI*diam;
    params.dimmbase = 2*M_PI*dbase;
    params.PIx2lambda0 = 2*M_PI*lambda0*MKM2CM;
    params.phi = phi;
    params.re_sfun = re_sfun;
    params.im_sfun = im_sfun;
    vector<double> result( z.size() );
    for( unsigned int i = 0; i < z.size(); i++ ) {
        params.altitude  = z(i)*KM2CM;
        gsl_integration_qagiu( &F, 0, 0, EPS, WORKSPACE, w, &result(i), &error);
    }
    gsl_integration_workspace_free( w );
    return result*DIMMCONST;
}
