/*
    $Id: weif.h 101 2010-03-02 08:36:15Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _WEIF_H
#define _WEIF_H

#include <string>
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>

#include "vector_generator.h"

using namespace boost::numeric::ublas;

struct Spectral_data {
    std::string response;
    std::string weights;
    std::string spectra;
    std::string ccdresponse;
    Spectral_data( const std::string& r, const std::string& w, const std::string& s, const std::string& c)
        : response( r ), weights( w ), spectra( s ), ccdresponse(c) {}
};

class Spline;

typedef struct {
    double diameter1; 
    double diameter2;
    double diameter3;
    double diameter4;
    double PIx2lambda0;
    Spline *re_sfun;
    Spline *im_sfun;
    double altitude;
} mass_params;

typedef struct {
    double aperture;
    double dimmbase;
    double PIx2lambda0;
    double phi;
    Spline *re_sfun;
    Spline *im_sfun;
    double altitude;
} dimm_params;

class Weights {
    unsigned int  mass_nchan;
    unsigned int  weights_number;
    vector<double> zgrid;           /**< vector of altitudes: z-grid [km] */
    matrix<double> ws;              /**< matrix of weights ws[aper][z] */
    vector<double> eps_inn;         /**< real inner diameters of apertures in [cm] */
    vector<double> eps_out;         /**< real outer diameters of apertures in [cm] */
    double leff;                    /**< effective wavelength of weight func. [mkm] */
    double lblue;                   /**< 50\% level wavelength at the blue side of SED [mkm] */
    double lred;                    /**< 50\% level wavelength at the red side of SED [mkm] */
    double dimm_aperture;
    double dimm_base;
public:
    Weights( int, const vector<double>& );
    ~Weights() {}
    const matrix<double>& weights() const { return ws; }
    const vector<double>& grid() const { return zgrid; }
    unsigned int apertures_number() const { return mass_nchan; }
    unsigned int weight_functions_number() const { return weights_number; }
    double lambda_eff() const { return leff ;}
    void interpolate( const Weights& refwf );
    void calculation( const Spectral_data&, const vector<double>&, const vector<double>&, double, double );
    void write( const std::string& ) const ;
private:
    void calc( vector<double>&, vector<double>&, vector<double>&, vector<double>& );
    int calcleff( vector<double>& lambda, vector<double>& edist );
    void photon_distr( vector<double>&, vector<double>&, vector<double>&, vector<double>& );
    void read_crv( const std::string&, vector<double>&, vector<double>& ) ;
    void spectral_band_fft( vector<double>&, vector<double>&, int, Spline*&, Spline*& ) ;
    vector<double> mass_weight( const vector<double>&, double, double, double, double, double, Spline*, Spline* );
    vector<double> dimm_weight( const vector<double>&, double, double, double, double, Spline*, Spline* );
};

    template<class V> vector< typename V::value_type > grid_generate( const V& z ) {
        typedef V vector_type;
        typedef typename vector_type::value_type value_type;
        typedef typename vector_type::size_type size_type;
        typedef vector< value_type > return_vector_type;
        size_type n = z.size();
        const int z0 = 0;
        const int zmax = 1;
        const int dz = 2;
        const int dzmin = 3;
        return_vector_type h = scalar_vector<double>(1,z[z0]);
        if( n == 4 && z[zmax]> z[dz] && z[dzmin] != 0 ) {
            for( unsigned int i = 0; h(i) < z[zmax] ; i++ ) {
                h.resize(h.size()+1);
                h(i+1) = h(i) + std::max(0.5*h(i)*z[dz],z[dzmin]);
            }
        }
        if( n == 4 && z[zmax] > z[dz] && z[dzmin] == 0 ) {
            for( unsigned int i = 0; h(i) < z[zmax] ; i++ ) {
                h.resize(h.size()+1);
                h(i+1) = h(i) + z[dz];
            }
        }
        if( n > 1 && z[zmax]< z[dz] && z[dz] < z[dzmin] ) {
            h.resize(n);
            for( unsigned int i = 0; i < n ; i++ ) h(i) = z[i];
        }
        return h;
    }

    struct linear_generator {
        typedef double result_type;
        typedef size_t size_type;

        linear_generator( result_type start, result_type step );
        result_type operator()( size_type i ) const;
        size_type size( result_type max ) const;
    private:
        result_type start_,step_;
    };

    struct lg_generator {
        typedef double result_type;
        typedef size_t size_type;

        lg_generator( result_type start, result_type lstep, result_type gstep );
        result_type operator()( size_type i ) const;
        size_type size( result_type max ) const;
    private:
        result_type start_,t_,q_;
        size_type max_;
    };

#endif
