/*
    $Id: worksheet.cpp 117 2010-03-19 12:19:16Z matwey $
    Copyright (C) 2009  Sternberg Astronomical Institute, MSU

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <cmath> // for std::cos
#include <iostream>

#include <boost/algorithm/string.hpp>
#include <boost/bind.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/iterator.hpp>
#include <boost/iterator/filter_iterator.hpp>
#include <boost/format.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/numeric/ublas/vector.hpp>

#include <yalpa/sun.h>
#include <yalpa/sidtime.h>
#include <yalpa/spheric.h>

#include "atmos.h"
#include "average.h"
#include "dimm.h"
#include "weif.h" // for wf::*_generator
#include "worksheet.h"

#include "photometry.h"

namespace {
	template<class Iterator> inline double integral_mean(Iterator begin, Iterator end) {
		return average::average( boost::make_filter_iterator<is_positive_number>( begin, end ),
			boost::make_filter_iterator<is_positive_number>( end, end ), average::mean< double >() );
	}
	template<class Iterator> inline double integral_error(Iterator begin, Iterator end, double mean) {
		return average::average( boost::make_filter_iterator<is_positive_number>( begin, end ),
			boost::make_filter_iterator<is_positive_number>( end, end ), average::error_of_mean< double >( mean ) );
	}
};

struct worksheet::profile_grid {
	boost::numeric::ublas::vector<double> vec;

	profile_grid( const boost::numeric::ublas::vector<double>& vec ): vec(vec) {}
};

class worksheet::object:
	public local_context< worksheet > {
public:
	typedef context_type::input::time_type time_type;
	typedef context_type::spectral_factory_type::weight_type weight_type;
private:
	std::string m_spectral;
	yalpa::equatorial<double> m_coords;
public:
	object( const yalpa::equatorial<double> ecoords, const std::string& spectral, context_reference_type ctx ): m_coords(ecoords), m_spectral(spectral), local_context_type( ctx ) {}

	double airmass( const time_type& time ) const;
	const weight_type& spectral();
	const std::string& sclass() const { return m_spectral; }
};

worksheet::input::measurement_accumulator::~measurement_accumulator() {
}

class worksheet::input::background_accumulator:
	public measurement_accumulator {
private:
	std::auto_ptr<background_model> m_model;
public:
	background_accumulator( const boost::posix_time::ptime& time, context_reference_type ctx ):
		measurement_accumulator( time, ctx ),
		m_model(ctx.m_bfactory.create_new_model(time) )
	{
		if( !m_model.get() )
			throw std::logic_error("There is not any background model");
	};
	virtual ~background_accumulator() {};
	virtual void accumulate(  const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) {
		if( context().m_bfactory.filter( time, mean, vars ) ) {
			if( m_model.get() ) {
				m_model->accumulate( time, mean );
			}
		}
	}
	virtual void dimm_accumulate( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars ) {
	}
	virtual void flush() {
		context().m_file_output.background( starttime() );

		if( m_model.get() && m_model->size() > 0 ) {
			context().m_bfactory.handle_new_model( m_model.release() );
			context().m_file_output.fluxes( starttime(), context().m_bfactory.background( starttime() ) );
		}
	}
};
class worksheet::input::normal_accumulator:
	public measurement_accumulator {
public:
	typedef context_type::input::size_type size_type;
private:
	boost::posix_time::ptime m_accumulation_begin;
	boost::posix_time::ptime m_accumulation_end;
	double m_mean_mz;
	size_type m_num;

	dimm_storage m_dimm;

	//FIXME: use params proxy lazy initialization
	const ublas::vector< double > m_deadtime;
	const ublas::vector< double > m_scatter;
	const ublas::vector< double > m_npoisson;

	indices_storage m_indices;
public:
	normal_accumulator( const boost::posix_time::ptime& time, context_reference_type ctx ):
		measurement_accumulator( time, ctx ),
		m_mean_mz(0.0),
		m_num(0),
		m_deadtime( ctx.m_pproxy.nonlinearity() ),
		m_scatter( ctx.m_pproxy.scatter() ),
		m_npoisson( ctx.m_pproxy.nonpoisson() ) {
			
		m_indices.add_filter( indices_threshold_filter(2.0) );
	}
	virtual ~normal_accumulator() {};
	virtual void accumulate( const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) {
		if( m_accumulation_begin == boost::posix_time::not_a_date_time )
			m_accumulation_begin = time;
		m_accumulation_end = time;

		m_num++;
		m_mean_mz = m_mean_mz * ( double(m_num-1) / double(m_num) ) + context().curobject().airmass( time ) * ( 1.0/m_num );

		indices::flux_type flux_corrected = indices::flux_nonlinearity_correction( mean, m_deadtime );
		m_indices.flux_filter( flux_corrected, true );

		boost::tuple< vars_type, rho1_type, rho2_type > vars_corrected = indices::vars_correction(
			flux_corrected,
			element_prod( vars, outer_prod( mean, mean ) ),
			element_prod( rho1, outer_prod( mean, mean ) ),
			element_prod( rho2, element_prod( mean, mean ) ), m_deadtime, m_npoisson );

		flux_corrected = indices::flux_background_correction( flux_corrected, m_scatter, context().m_bfactory.background( time ) );
			
		indices::indices_type indices;
		indices::indesis_type indesis;
		boost::tie(indices,indesis) = indices::calculate( flux_corrected, vars_corrected.get<0>(), vars_corrected.get<1>(), vars_corrected.get<2>() );
		
		m_indices.push_back( time, flux_corrected, indices, indesis );
	}
	virtual void dimm_accumulate( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars ) {
		if( m_accumulation_begin == boost::posix_time::not_a_date_time )
			m_accumulation_begin = time;
		m_accumulation_end = time;
		
		m_dimm.push_back( time, framenum, left, right, dimm::correction( xvars ), dimm::correction( yvars ) );
	}
	virtual void flush() {
		const boost::posix_time::ptime& meantime = m_accumulation_begin + (m_accumulation_end-m_accumulation_begin)/2;

		typedef context_type::integrals_factory_type integrals_factory_type;
		typedef integrals_factory_type::functor_type integrals_functor_type;
		typedef integrals_functor_type::result_type  integrals_result_type;
		typedef restoration_functor                   restoration_functor_type;
		typedef restoration_functor_type::result_type restoration_result_type;

		context().m_file_output.normal( starttime(), m_mean_mz );
		
		if( m_indices.size() > 1 ) {
			double dimm_seeing = 0.0;
			double dimm_turbulence = 0.0;
			double dimm_seeing_err = 0.0;
			double dimm_turbulence_err = 0.0;
			
			if( m_dimm.size() > 1 ) {
				m_dimm.correct_low_vars();

				dimm::dimm_vars_type xvars = average::average( m_dimm.xvars().begin(), m_dimm.xvars().end(), average::mean< dimm_vars_type >() );
				dimm::dimm_vars_type yvars = average::average( m_dimm.yvars().begin(), m_dimm.yvars().end(), average::mean< dimm_vars_type >() );
				dimm::dimm_vars_type xerrs = average::average( m_dimm.xvars().begin(), m_dimm.xvars().end(), average::error_of_mean< dimm_vars_type >(xvars) );
				dimm::dimm_vars_type yerrs = average::average( m_dimm.yvars().begin(), m_dimm.yvars().end(), average::error_of_mean< dimm_vars_type >(yvars) );

				double Jlong = dimm::turbulence( context().m_pproxy.dimm_scale(), context().m_pproxy.dimm_Klongitudinal(), xvars(1) );
				double Jlong_err = dimm::turbulence( context().m_pproxy.dimm_scale(), context().m_pproxy.dimm_Klongitudinal(), xerrs(1) );
				double Jtrans = dimm::turbulence( context().m_pproxy.dimm_scale(), context().m_pproxy.dimm_Ktransversal(),  yvars(1) );
				double Jtrans_err = dimm::turbulence( context().m_pproxy.dimm_scale(), context().m_pproxy.dimm_Ktransversal(),  yerrs(1) );

				dimm_turbulence = 0.5*( Jlong + Jtrans ) / m_mean_mz;
				dimm_turbulence_err = 0.5 / m_mean_mz * std::sqrt( Jlong_err*Jlong_err + Jtrans_err*Jtrans_err );
				dimm_seeing = dimm::seeing( dimm_turbulence );
				dimm_seeing_err = 0.6 * dimm_seeing / dimm_turbulence * dimm_turbulence_err;
			}
			
			indices::flux_type flux_avg = average::average( m_indices.flux().begin(), m_indices.flux().end(), average::mean< indices::flux_type >() );
			indices::flux_type flux_err = average::average( m_indices.flux().begin(), m_indices.flux().end(), average::error_of_mean< indices::flux_type >(flux_avg) );
			indices::indices_type indices_avg = average::average( m_indices.indices().begin(), m_indices.indices().end(), average::mean<indices::indices_type>() );
			indices::indices_type indices_err = average::average( m_indices.indices().begin(), m_indices.indices().end(), average::error_of_mean<indices::indices_type>(indices_avg) );
			indices::indesis_type indesis_avg = average::average( m_indices.indesis().begin(), m_indices.indesis().end(), average::mean<indices::indesis_type>() );
			indices::indesis_type indesis_err = average::average( m_indices.indesis().begin(), m_indices.indesis().end(), average::error_of_mean<indices::indesis_type>(indesis_avg) );
			
			context().m_file_output.indices( meantime, indices_avg, indices_err, indesis_avg, indesis_err );
			context().m_file_output.fluxes( meantime, flux_avg, flux_err );

			integrals_functor_type ifun = context().m_ifactory.get_functor(
				context().curobject().sclass(),
				context().curobject().spectral() );

			integrals_storage integrals;
			for( indices_storage::const_iterator it = m_indices.begin(); it != m_indices.end(); ++it ) {
				integrals.push_back( ifun( it->indices, it->indesis, m_mean_mz ) );
			}

			double free_seeing_avg       = integral_mean( integrals.free_seeing().begin(),       integrals.free_seeing().end() );
			double isoplanatic_angle_avg = integral_mean( integrals.isoplanatic_angle().begin(), integrals.isoplanatic_angle().end() );
			double altitude_eff_avg      = integral_mean( integrals.altitude_eff().begin(),      integrals.altitude_eff().end() );
			double time_constant_avg     = integral_mean( integrals.time_constant().begin(),     integrals.time_constant().end() );
			double moment_0_avg          = integral_mean( integrals.moment_0().begin(),          integrals.moment_0().end() );
			double moment_2_avg          = integral_mean( integrals.moment_2().begin(),          integrals.moment_2().end() );
			double free_seeing_err       = integral_error( integrals.free_seeing().begin(),       integrals.free_seeing().end(),       free_seeing_avg );
			double isoplanatic_angle_err = integral_error( integrals.isoplanatic_angle().begin(), integrals.isoplanatic_angle().end(), isoplanatic_angle_avg );
			double altitude_eff_err      = integral_error( integrals.altitude_eff().begin(),      integrals.altitude_eff().end(),      altitude_eff_avg );
			double time_constant_err     = integral_error( integrals.time_constant().begin(),     integrals.time_constant().end(),     time_constant_avg );
			double moment_0_err          = integral_error( integrals.moment_0().begin(),          integrals.moment_0().end(),          moment_0_avg );
			double moment_2_err          = integral_error( integrals.moment_2().begin(),          integrals.moment_2().end(),          moment_2_avg );

			context().m_file_output.atmosphere( meantime,
				free_seeing_avg, free_seeing_err,
				dimm_seeing, dimm_seeing_err,
				dimm_turbulence, dimm_turbulence_err,
				isoplanatic_angle_avg, isoplanatic_angle_err,
				altitude_eff_avg, altitude_eff_err,
				time_constant_avg, time_constant_err,
				moment_0_avg, moment_0_err,
				moment_2_avg, moment_2_err );

			/* mixture */
			typedef vector< double > mixture_type;
			std::vector< mixture_type > mixture_samples;
			mixture_samples.reserve( m_indices.size() );

			vector< double > grid( context().m_va["profile-grid"].as< profile_grid >().vec );
			
			if( m_dimm.size() > 1 ) {
				double scale2 = context().m_pproxy.dimm_scale(); scale2*=scale2;
				dimm_storage::const_iterator dimm_current = m_dimm.begin();
				dimm_storage::const_iterator dimm_next = dimm_current + 1;
				indices_storage::const_iterator it = m_indices.begin();
				while( dimm_current != m_dimm.end() ) {
					while( it != m_indices.end() ) {
						if( dimm_next != m_dimm.end() && dimm_next->timestamp <= it->timestamp )
							break;
						
						mixture_type mixture_sample( 12 );
						project( mixture_sample, range(0,10) ) = restoration_functor::flat_reorder( it->indices );
						mixture_sample(10) = scale2 * dimm_current->xvars(1) * 1e+11;
						mixture_sample(11) = scale2 * dimm_current->yvars(1) * 1e+11;
						++it;
						mixture_samples.push_back( mixture_sample );
					}
					dimm_current = dimm_next;
					++dimm_next;
				}
				vector< double > newgrid( grid.size() + 1 );
				newgrid(0) = 0;
				subrange( newgrid, 1, grid.size()+1 ) = grid;
				grid = newgrid;
			} else {
				for( indices_storage::const_iterator it = m_indices.begin(); it != m_indices.end(); ++it ){
					mixture_samples.push_back( restoration_functor::flat_reorder( it->indices ) );
				}
			}
			/* covariate mixture */
			mixture_type mixture_avg = average::average( mixture_samples.begin(), mixture_samples.end(), average::mean< mixture_type >() );
			symmetric_matrix< double > mixture_covariance = average::average( mixture_samples.begin(), mixture_samples.end(), average::covariance< mixture_type >(mixture_avg) );
			triangular_adaptor< symmetric_matrix< double >, strict_lower > ( mixture_covariance ) *= 0.8;
			/* prepare grid */
				
			restoration_functor_type rfun(
				grid,
				mixture_covariance,
				m_mean_mz,
				context().curobject().spectral()
				);

			profile_storage profile;
			for( std::vector< mixture_type >::const_iterator it = mixture_samples.begin(); it != mixture_samples.end(); ++it ) {
				profile.push_back( rfun( *it ) );
			}

			double chi2_avg = average::average( profile.chi2().begin(), profile.chi2().end(), average::mean< double >() );
			vector< double > cn2_avg = average::average( profile.cn2().begin(), profile.cn2().end(), average::mean< vector< double > >() );
			vector< double > cn2_err = average::average( profile.cn2().begin(), profile.cn2().end(), average::error_of_mean< vector< double > >( cn2_avg ) );
			vector< double > residual_avg = average::average( profile.residual().begin(), profile.residual().end(), average::mean< vector< double > >() );

			context().m_file_output.profile( meantime, "X", rfun.seeing( cn2_avg ), chi2_avg, cn2_avg, cn2_err, grid );
			context().m_file_output.residual( meantime, "X", element_div( rfun.indices_residual( residual_avg ), mixture_avg ) );
		}

		m_accumulation_begin = m_accumulation_end = boost::posix_time::ptime();
		m_mean_mz = 0;
		m_num = 0;
	}
};

double worksheet::object::airmass( const time_type& time ) const {
	yalpa::horizon<double> hc( yalpa::make_horizon_helper( m_coords, context().m_pproxy.site(), yalpa::sidereal_time(time) ) );
	double secz1 = 1.0/( std::cos(hc.coaltitude()) );
	double secz2 = secz1 - 1.0;
	return secz1 - secz2*( 0.0018167 + secz2*( 0.002875 + 0.0008083*secz2 ) );
}
const worksheet::object::weight_type& worksheet::object::spectral() {
	return context().m_sfactory.spectral(
		m_spectral,
		context().m_pproxy.inner_diam(),
		context().m_pproxy.outer_diam(),
		context().m_pproxy.dimm_aperture_base(),
		context().m_pproxy.dimm_aperture_size() );
}

void validate(boost::any& v, const std::vector<std::string>& values, worksheet::profile_grid* target_type, int) {
	namespace po = boost::program_options;
	using namespace boost::numeric;

	typedef std::vector< std::string > split_vector_type;

	std::string s = po::validators::get_single_string(values);
	split_vector_type split_vec;

	try{
		boost::trim_if( s, boost::is_space() );
		boost::split( split_vec, s, boost::is_any_of(", "), boost::token_compress_on );
		if( split_vec.size() == 4 ){
			double start = boost::lexical_cast<double>( split_vec[0] );
			double stop  = boost::lexical_cast<double>( split_vec[1] );
			double step  = boost::lexical_cast<double>( split_vec[2] );
			double step2 = boost::lexical_cast<double>( split_vec[3] );
			if( step2 == 0.0 ){
				linear_generator gen( start, step );
				v = boost::any( worksheet::profile_grid(vector_generator< linear_generator > ( gen, gen.size( stop ) )) );
			} else {
				lg_generator gen( start, step2, step );
				v = boost::any( worksheet::profile_grid(vector_generator< lg_generator > ( gen, gen.size( stop ) )) );
			}
		} else {
			ublas::vector<double> vec( split_vec.size() );
			for( split_vector_type::size_type it = 0; it < split_vec.size(); ++it ) {
				vec(it) = boost::lexical_cast<double>( split_vec[it] );
			}
			v = boost::any( worksheet::profile_grid(vec) );
		}
	} catch( const std::exception& e ) {
		throw po::validation_error("invalid value");
	}
}

int worksheet::parse_program_options( int argc, char** argv, boost::program_options::variables_map& va ) {
	namespace po = boost::program_options;

	po::options_description desc("Allowed options");
	desc.add_options()
		( "help,h",    "produce help message" )
		( "version,v", "produce version message" )
		( "accum-time,a",    po::value<double>()->default_value(60.0), "force accumulation time" )
		( "background-threshold,b", po::value<double>()->default_value(4.0), "set background threshold" )
		( "profile-grid,Z",  po::value<profile_grid>()->default_value( profile_grid( vector_generator< lg_generator >( lg_generator( 0.5, 1.0, 2.0 ),6 ) ), "0.5,1,2,4,8,16" ), "set for profile grid generator" )
		( "mass-response,r", po::value<std::string>()->default_value("mass_maid.crv"), "file with mass detector response" )
		( "ccd-response,c",  po::value<std::string>()->default_value("ccd.crv"), "file with ccd response" )
		( "dump-weight,w",   "dump weight functions to the files")
		( "spectra-dir,s",   po::value<std::string>()->default_value("spectra"), "spectra-files directory" );

	try {
		po::store( po::parse_command_line( argc, argv, desc ), va );
		po::notify( va );
		if( va.count("version") ) {
			std::cerr << (boost::format("%1% %2%") % PACKAGE % VERSION) << std::endl;
			std::cerr << "Copyright (C) 2009  Sternberg Astronomical Institute, MSU" << std::endl;
			std::cerr << "This is free software; see the source for copying conditions.  There is NO" << std::endl;
			std::cerr << "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE." << std::endl;
			return 1;
		}
		if( va.count("help") ) {
			std::cerr << desc << std::endl;
			return 1;
		}
	} catch ( const po::error& e ) {
		std::cerr << e.what() << std::endl;
		std::cerr << desc << std::endl;
		return 1;
	}
	return 0;
}
background_model* worksheet::select_background_model( const boost::posix_time::ptime& time ) const {
	double alt = sun_altitude( time ) * 180 / M_PI; // degrees
	
	if( alt > -5.0 )
		return new background_daytime_model( time );
	else if( alt > -12.0 )
		return new background_twilight_model( time, boost::bind(std::mem_fun(&worksheet::sun_altitude), this, _1) );
	// else
	return new background_night_model( time );
}
double worksheet::sun_altitude( const boost::posix_time::ptime& time ) const {
	yalpa::ecliptic<double> sun = yalpa::sun_position( time );
	yalpa::horizon<double> sun_hz( make_horizon_helper( sun, m_pproxy.site(), yalpa::sidereal_time(time) ) );
	return 0.5 * M_PI - sun_hz.coaltitude();
}
worksheet::worksheet( const boost::program_options::variables_map& va ):
	m_va(va),
	m_pcontainer(),
	m_pproxy( m_pcontainer ),
	m_sfactory( boost::filesystem::path( va["spectra-dir"].as<std::string>() ), boost::filesystem::path( va["mass-response"].as<std::string>() ), boost::filesystem::path( va["ccd-response"].as<std::string>() ) ),
	m_bfactory( boost::bind( std::mem_fun( &worksheet::select_background_model ), this, _1) ),
	m_file_input( std::cin, *this ),
	m_file_output( std::cout )
{

	m_bfactory.add_filter( background_nonpoisson_filter(
				va["background-threshold"].as<double>(),
				boost::bind(std::mem_fun(&worksheet::nonpoisson), this, 3),
				boost::bind(std::mem_fun(&worksheet::numexp), this) ) );
}

void worksheet::pre() {
	/* do nothing */
	m_file_output.comment((boost::format("%1% %2%") % PACKAGE % VERSION).str());
}
void worksheet::run() {
	bool ret = true;
	while( ret ) {
		try {
			ret &= m_file_input.shift();
		} catch ( const std::exception& e ) {
			std::cerr << m_file_input.line() << ": " << e.what() << std::endl;
		}
	}
}
void worksheet::post() {
	if( m_va.count("dump-weight") ) {
		for( spectral_factory_type::const_iterator it = m_sfactory.begin(); it != m_sfactory.end(); ++it ) {
			it->second.write( ( boost::format("weight-%1%.wf") % it->first ).str() );
		}
	}
}
worksheet::~worksheet() {
	
}

void worksheet::input::on_comment( const std::string& line ) {
	/* do nothing */
}
void worksheet::input::on_parameter( const time_type& time, const std::string& key, const std::string& value ) {
	context().m_pcontainer[key] = value;

	context().m_file_output.parameter( time, key, value );
}
void worksheet::input::on_object( const time_type& time, const object_type& obj ) {
	if( accumulator_type< dummy_accumulator >() ) {
		context().m_curobject = std::auto_ptr< context_type::object >( new context_type::object(obj.coordinates,obj.sclass,context()) );

		context().m_file_output.objectinfo( time, obj );
	} else {
		unshift();
		switch_accumulator<dummy_accumulator>( time );
	}
}
void worksheet::input::on_background( const time_type& time ) {
	if( accumulator_type< dummy_accumulator >() ) {
		switch_accumulator<background_accumulator>( time );
	} else {
		unshift();
		switch_accumulator<dummy_accumulator>( time );
	}
}
void worksheet::input::on_normal( const time_type& time ) {
	if( accumulator_type< dummy_accumulator >() ) {
		switch_accumulator<normal_accumulator>( time );
	} else {
		if( !accumulator_type< normal_accumulator >() || (( time - m_maccum->starttime() ).total_seconds() >= context().m_va["accum-time"].as<double>() ) ) {
			unshift();
			switch_accumulator<dummy_accumulator>( time );
		}
	}
}
void worksheet::input::on_flux( const time_type& time ) {
	switch_accumulator<dummy_accumulator>( time ); // dummy no throw
}
void worksheet::input::on_measurements( const time_type& time, const flux_type& mean, const vars_type& vars, const rho1_type& rho1, const rho2_type& rho2 ) {
	m_maccum->accumulate( time, mean, vars, rho1, rho2 );
}
void worksheet::input::on_dimm( const time_type& time, unsigned int framenum, const dimm_flux_type& left, const dimm_flux_type& right, const dimm_vars_type& xvars, const dimm_flux_type& yvars ) {
	m_maccum->dimm_accumulate( time, framenum, left, right, xvars, yvars );
}


